# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import signal
import sys
from typing import Optional

from logger.config import get_config
from logger.constants import Constants
from logger.error import TerminateException, raise_terminate_exception
from logger.log import LoggingService
from logger.measure.controller import Controller
from logger.process import prepare_process
from logger.utils.env import parse_int
from logger.utils.exclusion_control import ExclusionControl, LockError

EX_OK = 0
EX_FAILURE = 1


def main():  # noqa: C901
    # 外部シグナルで終了する際にはTerminateExceptionをあげる
    signal.signal(signal.SIGTERM, raise_terminate_exception)
    signal.signal(signal.SIGINT, raise_terminate_exception)

    prepare_process()

    logging_service = LoggingService.get_instance()
    lock = ExclusionControl("/tmp/measure.lock")
    # finallyで安全に扱うため、事前に変数を定義する
    controller: Optional[Controller] = None

    try:
        logging_service.start()
        logger = logging.getLogger(__name__)
        # killするときようにプロセスIDを出力
        logger.debug(f"process id: {os.getpid()}")

        # 同時に計測が走らないように排他制御を行う
        try:
            lock.acquire()
        except LockError:
            logger.error("Another measurement is running, can't start measurement")
            return EX_FAILURE

        args = sys.argv
        # 引数が2個（1番目はプログラム名）でなかった場合は開始しない
        if len(args) != 2:
            logger.error(
                'Invalid arguments length. The argument must be "auto", "manual", or an integer.'
            )
            return EX_FAILURE

        if args[1] in [Constants.START_TYPE_AUTO, Constants.START_TYPE_MANUAL]:
            start_type = args[1]
            meas_time_sec_arg = None
        # 引数がautoかmanualでなかった場合のみ引数のパースを行う
        else:
            start_type = Constants.START_TYPE_SCHEDULE
            meas_time_sec_arg = parse_int(args[1])
            if meas_time_sec_arg is None:
                logger.error("Measurement time argument must be an integer")
                return EX_FAILURE

        config = get_config({"measure_time_sec": meas_time_sec_arg})

        if config is None:
            logger.error("Config file has wrong value, can't start measurement")
            return EX_FAILURE

        controller = Controller(config=config, start_type=start_type)
        try:
            controller.init()
        # 実行中に中断シグナルを受け取った場合はログ出力をしない
        except TerminateException:
            return EX_OK
        except Exception:
            logger.exception("Failed to initialize measurement")
            logger.error("Can't start measurement")
            return EX_FAILURE

        try:
            controller.start()
        # 実行中に中断シグナルを受け取った場合はログ出力をしない
        except TerminateException:
            return EX_OK
        except Exception:
            logger.exception("Failed to measure")
            logger.error("Can't continue measurement")
            return EX_FAILURE

        return EX_OK

    except TerminateException:
        pass
    except Exception:
        logger.exception("Unknown error occurred")
        return EX_FAILURE
    finally:
        output_dir: Optional[str] = None
        if controller:
            output_dir = controller.get_output_dir()
            # タイマー実行の時にはすでに計測終了しているはずなので、terminateはTrueで良い
            controller.stop(terminate=True)
        logging_service.close(output_dir)

        lock.release()


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)
