# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import time
from dataclasses import dataclass
from typing import List, Optional, Union

from serial import Serial


@dataclass
class WaitCommand:
    """Serialにメッセージを送る際一定時間停止することを表すコマンド"""

    sec: float


Command = List[int]

# コマンドリスト型
CommandList = List[Union[WaitCommand, Command]]


class Comm:
    """シリアルとの通信クラス"""

    DEFAULT_TIMEOUT = 3
    DEFAULT_READ_BYTES = 4096

    def __init__(self):
        self._connection: Optional[Serial] = None
        self._logger = logging.getLogger(__name__)

    def open(
        self,
        port: str,
        baud: int,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
    ) -> None:
        """シリアルとのコネクションを開くメソッド

        Args:
            port (str): 通信対象のポート
            baud (int): 通信速度
            timeout (Optional[float], optional): 接続のタイムアウト時間。Noneを渡した場合、タイムアウトしなくなる。 デフォルトはConstants.COMM_TIMEOUT_DEFAULT(3秒)。
        """

        self._logger.debug(f"Open: {port} {baud} {timeout}")
        self._connection = Serial(port, baud, timeout=timeout)

    def send(self, commands: CommandList) -> List[int]:
        """コマンドを送って結果を受け取るメソッド

        Args:
            commands (CommandList): デバイスに送信するコマンド、WaitCommandを送った場合その時間だけ待機する

        Raises:
            TimeoutError: タイムアウト内にデータが帰ってこなかった場合
            AssertionError: openメソッド実行前に呼び出した場合

        Returns:
            List[int]: 受け取った結果
        """

        assert self._connection

        result = []

        for command in commands:
            self._logger.debug(f"Send command: {command}")
            if isinstance(command, WaitCommand):
                time.sleep(command.sec)
            else:
                self._connection.write(bytes(command[1:]))
                self._connection.flush()

                # データ受信
                if command[0] > 0:
                    read_result = self.read(command[0])
                    result.extend(read_result)

        return result

    def read(self, length: int) -> List[int]:
        """指定した長さのデータを読み取るメソッド

        Args:
            length (Optional[int]): 読み取りたいデータ長、未指定の場合は一回に読み取れたデータをそのまま返す

        Raises:
            TimeoutError: タイムアウト内にデータが帰ってこなかった場合
            AssertionError: openメソッド実行前に呼び出した場合

        Returns:
            List[int]: intの配列に変換した受け取ったデータ
        """
        return list(self.read_bytes(length))

    def read_bytes(self, length: int) -> bytes:
        """指定した長さのデータをバイトで読み取るメソッド

        Args:
            length (Optional[int]): 読み取りたいデータ長、未指定の場合は一回に読み取れたデータをそのまま返す

        Raises:
            TimeoutError: タイムアウト内にデータが帰ってこなかった場合
            AssertionError: openメソッド実行前に呼び出した場合

        Returns:
            bytes: 読み取ったバイト列
        """
        assert self._connection

        result = bytes()

        while length > 0:
            read_length = (
                length if length < Comm.DEFAULT_READ_BYTES else Comm.DEFAULT_READ_BYTES
            )
            if read_length > 0:
                received = self._connection.read(read_length)
                read_length = len(received)

                # timeout 判定
                if read_length == 0:
                    raise TimeoutError("Read timeout occurred")

                # データが分割されることもあるため list を拡張・保持する
                # datalen を超えるデータは保持しない
                result += received[0:length]
                length -= read_length

        return result

    def close(self) -> None:
        """デバイスとの通信を終了するメソッド"""
        if self._connection is None:
            return
        self._logger.debug("Close")
        self._connection.reset_input_buffer()
        self._connection.close()
        self._connection = None


def send(
    commands: CommandList,
    port: str,
    baud: int,
    timeout: Optional[float] = Comm.DEFAULT_TIMEOUT,
) -> List[int]:
    """通信の確立から終了まで含め、デバイスにコマンドを送り、メッセージを受け取る関数

    Args:
        commands (CommandList): デバイスに送信するコマンド、WaitCommandを送った場合その時間だけ待機する
        port (str): 通信対象のポート
        baud (int): 通信速度
        timeout (Optional[float], optional): 接続のタイムアウト時間。Noneを渡した場合、タイムアウトしなくなる。 デフォルトはConstants.COMM_TIMEOUT_DEFAULT(3秒)。

    Returns:
        List[int]: 読み取った結果
    """
    try:
        com = Comm()
        com.open(port, baud, timeout)
        result = com.send(commands)
    finally:
        try:
            com.close()
        except Exception:
            pass
    return result
