# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import logging.handlers
import multiprocessing
import os
import shutil
import signal
import sys
import traceback
from typing import Callable, Optional

from logger.constants import ConfigKeys, Constants
from logger.utils.dir import root_dir
from logger.utils.env import get_str


class LoggingClient:
    """他プロセスに渡してプロセスセーフなファイルロガーをセットアップするためのクラス

    Attributes:
        queue multiprocessing.Queue: リスナープロセスにログメッセージを送出するqueue
        configure Callable[[multiprocessing.Queue], Any]: プロセスで使用されるロガーにqueueハンドラーをセットアップする関数
    """

    def __init__(
        self,
        queue: multiprocessing.Queue,
        configure: Callable[[multiprocessing.Queue], None],
    ) -> None:
        self.queue = queue
        self.configure = configure


class _Formatter(logging.Formatter):
    def format(self, rec: logging.LogRecord) -> str:
        return "%-8s|%s %s: %s" % (
            rec.levelname,
            self.formatTime(rec),
            rec.name,
            rec.getMessage(),
        )


class LoggingService:
    """プロセスセーフなロギングを管理するクラス

    シングルトンとして振る舞い、get_instanceメソッドを通じてインスタンスを取得する
    """

    _instance: Optional["LoggingService"] = None
    LOG_LEVEL: str = get_str(ConfigKeys.LOG_LEVEL) if get_str(ConfigKeys.LOG_LEVEL) in Constants.LOG_LEVELS else "INFO"  # type: ignore

    def __new__(cls):
        raise NotImplementedError("Cannot Generate Instance By Constructor")

    @classmethod
    def __internal_new__(cls):
        instance = super().__new__(cls)
        instance.__init__()  # type: ignore
        return instance

    def __init__(self) -> None:
        self._listener_process: Optional[multiprocessing.Process] = None
        self._queue: multiprocessing.Queue = multiprocessing.Queue(-1)
        self._log_file: Optional[str] = None

    @classmethod
    def get_instance(cls) -> "LoggingService":
        """シングルトンのLoggingServiceインスタンスを取得するメソッド

        Returns:
            LoggingService:
        """
        if not cls._instance:
            cls._instance = cls.__internal_new__()
        return cls._instance  # type: ignore

    @staticmethod
    def _listen(queue: multiprocessing.Queue, log_file: str):
        # Ctrl^Cでの終了シグナルが子プロセスに伝搬するため無視する
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        root = logging.getLogger()
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(_Formatter())
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(_Formatter())
        root.addHandler(file_handler)
        root.addHandler(stream_handler)
        while True:
            try:
                record = queue.get()
                if record is None:
                    break
                logger = logging.getLogger(record.name)
                logger.handle(record)
            except Exception:
                print("Whoops! Problem:", file=sys.stderr)
                traceback.print_exc(file=sys.stderr)

    @staticmethod
    def _configure(queue: multiprocessing.Queue) -> None:
        h = logging.handlers.QueueHandler(queue)
        root = logging.getLogger()
        root.addHandler(h)
        root.setLevel(LoggingService.LOG_LEVEL)

    def start(self):
        """ログファイル出力を立ち上げてロギングを開始するメソッド"""
        # 複数回このメソッドが呼び出されても不整合が起きないよう、初回呼び出し時のみ実行
        if not self._listener_process:
            # メインプロセスからログ出力された時のために_configureを実行
            LoggingService._configure(self._queue)
            # 計測が開始できなかった場合に備え、開始時は一時ファイルにログ出力を行う
            self._log_file = f"{root_dir()}/tmp_log/{format(datetime.datetime.now(), '%Y%m%d_%H%M%S')}_measure.log"
            self._listener_process = multiprocessing.Process(
                target=LoggingService._listen, args=(self._queue, self._log_file)
            )
            self._listener_process.start()

    def get_logging_client(self) -> LoggingClient:
        """別プロセスでプロセスセーフなロギングを行うための変数を作るメソッド

        Returns:
            LoggingArgs:

        Raises:
            AssertionError: startメソッドが実行される前に呼び出した場合に発生する
        """
        assert self._listener_process
        return LoggingClient(queue=self._queue, configure=LoggingService._configure)

    def close(self, output_dir: Optional[str] = None) -> None:
        """ファイル出力プロセスを終了させるメソッド

        Args:
            output_dir (Optional[str], optional): ログファイルの移動先ディレクトリ。指定した場合は「指定ディレクトリ/measure.log」にログファイルが作成される
        """
        try:
            # 複数回このメソッドが呼び出されても良いよう、出力プロセスが存在する時のみprocess終了処理実行
            if self._listener_process is not None:
                self._queue.put_nowait(None)
                self._listener_process.join()
                self._listener_process = None
            # 計測完了時に呼び出された際には計測フォルダにログファイルを移動
            if output_dir and self._log_file and os.path.isdir(output_dir):
                shutil.move(self._log_file, f"{output_dir}/measure.log")
                self._log_file = None
        except Exception as e:
            # mainの終了時に呼ばれる処理なので、errorは無視する
            # ログ出力もできるかわからないためprintする
            print(f"logging service close failed with Error: {str(e)}")
