# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import math
import multiprocessing
import os
import platform
import threading
import time
from typing import List, Optional, Union

import serial.tools.list_ports

from logger.config import Config
from logger.constants import Constants, InfoKeys
from logger.measure.A342 import A342
from logger.measure.A352 import A352
from logger.measure.job import JobError
from logger.measure.sensor import Sensor, SensorDeviceError
from logger.utils.format import app_format
from logger.utils.timezone import get_time_zone


class Controller:
    def __init__(self, config: Config, start_type: str) -> None:
        self.config = config
        self._start_type = start_type
        self._sensors: List[Sensor] = []
        self._logger = logging.getLogger(__name__)
        self._stop_timer: Optional[threading.Timer] = None
        self._measuring = False
        self._error_queue: Optional[multiprocessing.Queue] = None
        self._start_time: Optional[datetime.datetime] = None
        self._end_time: Optional[datetime.datetime] = None

    def init(self) -> None:
        self._logger.info("Start initialization")
        # USBデバイスサーチ
        for port in Controller.list_comports():
            self._logger.debug(f"COM port found: {port}")
            # センサー初期化
            sensor = self._init_sensor(port)
            if sensor:
                self._sensors.append(sensor)
        if len(self._sensors) == 0:
            self._logger.error("No sensor found")
            raise ValueError("No Ready Sensor found")
        for sensor in self._sensors:
            sensor.init()

    def start(self) -> None:  # noqa: C901
        """計測を開始するメソッド

        Raises:
            self._error_queue.get: サブプロセスからのエラー発生
        """
        # スケジュール計測以外の場合のみ初回待機を行う
        if self._start_type != Constants.START_TYPE_SCHEDULE:
            self._logger.info(f"Initial wait for {self.config.initial_wait} sec")
            time.sleep(self.config.initial_wait)
        self._start_time = datetime.datetime.now()
        out_dir: str = self.get_output_dir()  # type: ignore
        os.makedirs(out_dir, exist_ok=True)
        # 計測時間が0より大きく設定されていたら、指定時間後に終了するようタイマーをセットする
        if self.config.measure_time_sec > 0:
            self._stop_timer = threading.Timer(self.config.measure_time_sec, self.stop)
            self._stop_timer.start()
        self._measuring = True
        # 計測ごとにerrorを受け取るqueueを作る
        self._error_queue = multiprocessing.Queue()
        self._logger.info("Start measurement")

        failed = set()
        for sensor in self._sensors:
            try:
                sensor.start(error_queue=self._error_queue, output_dir=out_dir)
            except Exception:
                self._logger.error(f"Sensor: {sensor} failed to start measurement")
                failed.add(f"{sensor.model}_{sensor.serial}")
        # 一つも計測が開始しなかった場合はstart処理をエラーで終了させる
        if len(failed) == len(self._sensors):
            self._logger.error("All sensors failed to start measurement")
            raise Exception("All sensors failed to start measurement")

        # 計測情報ファイル出力（後で計測時間を含めて上書きされる）
        self._output_info()

        # 計測中は1秒ごとに子プロセスのエラーがないか監視する
        while self._measuring:
            time.sleep(1)
            if self._error_queue and not self._error_queue.empty():
                error: JobError = self._error_queue.get()
                for sensor in self._sensors:
                    if sensor.model == error.model and sensor.serial == error.serial:
                        failed.add(f"{error.model}_{error.serial}")
                        self._logger.error(
                            f"Sensor: {sensor} got error during measurement"
                        )
                        try:
                            sensor.stop()
                        except Exception:
                            pass
                        break
                # 全てのセンサーの計測が失敗したらエラーを上げる
                if len(failed) == len(self._sensors):
                    self._logger.error("All sensors failed to measure")
                    raise Exception("All sensors failed to measure")

        # タイマーがセットされていてwhileループを抜けて正常終了した時のためにスレッドの終了を待つ
        if self._stop_timer:
            self._stop_timer.join()

    def stop(self, terminate: bool = False) -> None:
        """計測を終了するメソッド

        計測中以外に呼ばれた場合は何も行わない

        Args:
            terminate (bool, optional): SIG_TERMなどの中断によって呼ばれたか。 デフォルトは False.
        """
        # mainから計測中以外に呼ばれる場合があるためその際は何もせずにreturnする
        if not self._measuring:
            self._logger.debug("stop method called, but not measure")
            return
        try:
            # startのメソッドが実行中の場合、エラー待ち受けのループを停止するために計測中フラグをOFFにする
            self._measuring = False
            # エラーや終了シグナルで終了する際には非同期の終了タイマーを停止する
            if terminate and self._stop_timer:
                self._stop_timer.cancel()
                self._stop_timer = None

            for sensor in self._sensors:
                try:
                    sensor.stop()
                except Exception:
                    self._logger.error(f"Sensor: {sensor} failed to stop measurement")

            self._end_time = datetime.datetime.now()

            # 何によって計測が終了したかによってメッセージを出し分ける
            if terminate:
                self._logger.info("Terminate measurement")
            else:
                self._logger.info("Finish measurement")

            # 計測情報ファイル出力（計測後の上書き）
            self._output_info()

            # 計測で使ったqueueは消す
            if self._error_queue:
                self._error_queue.close()
                self._error_queue = None
        except Exception:
            # mainの終了時に呼ばれる処理なので、errorは無視してログ出力のみ行う
            self._logger.exception("An error occurred during stop measurement")

    def get_output_dir(self) -> Union[str, None]:
        if not self._start_time:
            return None
        return f"{self.config.output_path}/{self._start_time:%Y%m%d_%H%M%S}"

    @classmethod
    def list_comports(cls) -> List[str]:
        """通信可能なポートの一覧を取得するメソッド

        Returns:
            List[str]: 通信できるポートの一覧
        """
        result = []
        for com in serial.tools.list_ports.comports():
            if com.device.startswith("/dev/ttyAMA"):
                # ラズパイの GPIO ポートのため使用しない
                continue
            result.append(com.device)
        return sorted(result)

    def _init_sensor(self, port: str) -> Union[Sensor, None]:
        """センサーの初期化をするメソッド

        Args:
            port (str): センサーと通信を試みるポート

        Raises:
            SensorDeviceError: センサーが通信可能な状態でない場合

        Returns:
            Union[Sensor, None]: 初期化に成功したセンサー
        """
        # センサーとの初期通信
        try:
            sensor = self._create_sensor(port)
        except ConnectionRefusedError:
            self._logger.error(f"Failed to connect: {port}")
            return None

        # センサーの状態チェック
        try:
            # 状態チェック
            if not sensor.ready():
                raise SensorDeviceError()

            # シリアル読み込み
            sensor.load_serial()
            # ファームウェアバージョン読み込み
            sensor.load_firm_ver()
        except Exception:
            self._logger.error(f"Failed to initialize: {port}")
            return None

        return sensor

    def _create_sensor(self, port: str) -> Sensor:
        """センサー(具象クラス)を作成するメソッド

        Args:
            port (str): センサーとの通信を試みるポート

        Raises:
            ValueError: 不正なセンサーのproduct_idが取得された場合

        Returns:
            Sensor: 作成されたセンサー(具象クラス)
        """
        # product_idと通信成功したボーレートを取得
        product_id = Sensor.load_product_id(port, self.config.baud_rate)
        self._logger.info(
            f"Found product: {product_id} port={port} baud={self.config.baud_rate}"
        )
        if product_id.startswith("A352"):
            return A352(
                port=port,
                baud=self.config.baud_rate,
                product_id=product_id,
                sps=self.config.A352_sps,
                logger_id=self.config.logger_id,
                file_rotate_min=self.config.file_rotate_min,
            )
        if product_id.startswith("A342"):
            return A342(
                port=port,
                baud=self.config.baud_rate,
                product_id=product_id,
                logger_id=self.config.logger_id,
                file_rotate_min=self.config.file_rotate_min,
                mode=self.config.A342_mode,
                physical=self.config.A342_physical,
                rms_pp_interval=self.config.A342_rmspp_output_interval,
            )

        else:
            raise ValueError(f"Unknown PRODUCT_ID: {product_id}")

    def _output_info(self) -> None:
        """計測情報ファイルを出力する関数

        Raises:
            AssertionError: 計測開始前に呼ばれた場合（保存先ディレクトリが存在しない）
        """

        output_dir = self.get_output_dir()
        assert output_dir
        meas_time: Optional[int] = None
        if self._start_time and self._end_time:
            meas_time = math.floor((self._end_time - self._start_time).total_seconds())
        for sensor in self._sensors:
            info_text = (
                f"{InfoKeys.OS},{platform.platform()}\n"
                + f"{InfoKeys.LOGGER_VERSION},{Constants.LOGGER_VERSION}\n"
                + f"{InfoKeys.LOGGER_ID},{self.config.logger_id}\n"
                + f"{InfoKeys.START_TYPE},{self._start_type}\n"
                + f"{InfoKeys.TIME_ZONE},{get_time_zone()}\n"
                +
                # f-stringを入れ子にできないため、文字列連結する
                f"{InfoKeys.DATE},"
                + app_format(self._start_time, "%Y/%m/%d %H:%M:%S")
                + "\n"
                + f"{InfoKeys.MEASURE_TIME_SEC},{app_format(meas_time)}\n"
                + f"{InfoKeys.INITIAL_WAIT},{0 if self._start_type == Constants.START_TYPE_SCHEDULE else self.config.initial_wait}\n"
                + f"{InfoKeys.FILE_ROTATE_MIN},{self.config.file_rotate_min}\n"
                + f"{InfoKeys.BAUD_RATE},{self.config.baud_rate}\n"
                + f"{InfoKeys.PORT},{sensor.port}\n"
                + f"{InfoKeys.SENSOR},{sensor.model}\n"
                + f"{InfoKeys.PRODUCT_ID},{sensor.product_id}\n"
                + f"{InfoKeys.SERIAL_NO},{sensor.serial}\n"
                + f"{InfoKeys.FIRM_VER},{sensor.firm_ver:#06x}\n"
                + f"{InfoKeys.PHYSICAL},{sensor.physical}\n"
                + f"{InfoKeys.OUTPUT_TYPE},{sensor.mode}\n"
                + f"{InfoKeys.SPS},{sensor.sps}\n"
                + f"{InfoKeys.FILTER},{app_format(sensor.filter)}\n"
                + f"{InfoKeys.RMS_PP_INTERVAL},{app_format(sensor.rms_pp_interval)}\n"
            )

            # 計測ディレクトリのinfo.csvに書き込み
            with open(
                f"{output_dir}/{sensor.model}_{sensor.serial}_info.csv",
                mode="w",
                encoding="UTF-8",
                newline="\r\n",
            ) as f:
                f.write(info_text)
