# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import re
from typing import Any, List


class Validator:
    """バリデーション実行クラス"""

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)

    def _defined(self, value: Any, key: str) -> bool:
        """検証対象が存在するか（Noneではないか）を検証するメソッド

        Args:
            value (Any): 検証対象
            key (str): 検証対象の名前（失敗時の出力に利用）

        Returns:
            bool: 検証に成功したか
        """
        result = value is not None
        if not result:
            self.logger.error(f"{key} must be defined")
        return result

    def one_of(self, value: Any, key: str, list: List[Any]) -> bool:
        """検証対象が与えられた配列に含まれるかを検証するメソッド

        Args:
            value (Any): 検証対象
            key (str): 検証対象の名前（失敗時の出力に利用）
            list (List[Any]): 検証対象が含まれるべき配列

        Returns:
            bool: 検証に成功したか
        """

        result = self._defined(value=value, key=key) and value in list
        if not result:
            self.logger.error(f"{key} must be one of {','.join(map(str, list))}")
        return result

    def between(self, value: Any, key: str, min: Any, max: Any) -> bool:
        """検証対象が与えられた最小値と最大値の間に含まれるかを検証するメソッド

        Args:
            value (Any): 検証対象
            key (str): 検証対象の名前（失敗時の出力に利用）
            min (Any): 値の許容する最小値
            max (Any): 値の許容する最大値

        Returns:
            bool: 検証に成功したか
        """

        result = self._defined(value=value, key=key) and min <= value and value <= max
        if not result:
            self.logger.error(f"{key} must be between {min} and {max}")
        return result

    def match(self, value: Any, key: str, pattern: str) -> bool:
        """検証対象が与えられた正規表現にマッチするかを検証するメソッド

        Args:
            value (Any): 検証対象
            key (str): 検証対象の名前（失敗時の出力に利用）
            pattern (str): 検証する正規表現

        Returns:
            bool: 検証に成功したか
        """

        result = (
            self._defined(value=value, key=key) and re.match(pattern, value) is not None
        )
        if not result:
            self.logger.error(f"{key} must match pattern {pattern}")
        return result

    def exist_dir(self, path: Any, key: str) -> bool:
        """検証対象が実際に存在するディレクトリかを検証するメソッド

        Args:
            path (Any): 検証対象
            key (str): 検証対象の名前（失敗時の出力に利用）

        Returns:
            bool: 検証に成功したか
        """

        result = self._defined(value=path, key=key) and os.path.isdir(path)
        if not result:
            self.logger.error(f"{key}({path}) must be an existing directory")
        return result
