# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import multiprocessing
import os
from dataclasses import dataclass
from multiprocessing.sharedctypes import Synchronized
from typing import List

import pytest

from logger.measure.data import MeasureData
from logger.measure.job import ReaderArgs, WriterArgs, reader_job, writer_job


@dataclass
class ReaderArgsForTest(ReaderArgs):
    """テスト用に使うReaderArgs具象クラス"""

    def __post_init__(self):
        self._count = 1
        self._count_max = 4

    def format_packet(self, index: int, packet: List[int]) -> MeasureData:
        # パケット境界の修正がreader_jobで行われることを確認するためそちらを通過して不正なpacketが渡されたらエラーを挙げる
        if packet[0] != 0x80 or packet[-1] != 0x0D:
            raise ValueError("Invalid packet boundary")

        # A342のデータに合わせる
        count = self._count % self._count_max
        self._count += 1
        # 戻り値はダミー
        dummy = MeasureData(
            index=index, count=count, temperature=0, x=0, y=0, z=0, flag=0
        )
        return dummy

    def _gen_invalid_data(self, index: int, count: int) -> MeasureData:
        return MeasureData(
            index=index, count=count, temperature=0, x=0, y=0, z=0, flag=0
        )


@dataclass
class WriterArgsForTest(WriterArgs):
    def __post_init__(self):
        self._count = 0

    def get_file_name(self, start_time: datetime.datetime) -> str:
        # タイムスタンプで作られると外部環境に依存してやりづらいので内部indexをincrementしてファイル名をつける
        file_name = f"{self._count}.csv"
        self._count += 1
        return file_name


int_16bit_max = 2**16


class TestJob:
    class TestReaderJob:
        def test_read_job_success(self, mocker, mock_comm):
            # readループを1回で抜けれるようにmultiprocessing.Valueの戻り値をモックしておく
            mocker.patch(
                "multiprocessing.sharedctypes.Synchronized.value",
                side_effect=[True, False],
                new_callable=mocker.PropertyMock,
            )
            valid_packet = [0x80, 0, 0, 0x0D]
            mock_comm(
                send_return=[],
                read_return=[
                    *valid_packet,
                    *valid_packet,
                    *valid_packet,
                ],
            )
            queue = multiprocessing.Queue(-1)
            error_queue = multiprocessing.Queue(-1)
            measuring: Synchronized = multiprocessing.Value("b", True)
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=len(valid_packet),
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            reader_job(
                queue=queue, error_queue=error_queue, measuring=measuring, args=args
            )

            assert not queue.empty()
            records = queue.get()
            # 一回目の取得時は1秒分のレコードが入っている
            assert len(records) == args.record_per_sec
            records = queue.get()
            # 二回目の取得時はループを抜けているので残りが全て取得できる
            assert len(records) == 1
            records = queue.get()
            # ループを抜けたので終了フラグのNoneを返す
            assert records is None
            assert queue.empty()

        def test_read_job_success_with_invalid_packets(self, mocker, mock_comm):
            # readループを1回で抜けれるようにmultiprocessing.Valueの戻り値をモックしておく
            mocker.patch(
                "multiprocessing.sharedctypes.Synchronized.value",
                side_effect=[True, False],
                new_callable=mocker.PropertyMock,
            )
            valid_packet = [0x80, 0, 0, 0x0D]
            mock_comm(
                send_return=[],
                # 間に不正なパケットが入っていても正しく読み取れるか確認
                read_return=[
                    *valid_packet,
                    0,
                    1,
                    2,
                    4,
                    *valid_packet,
                    0x80,
                    0,
                ],
            )
            queue = multiprocessing.Queue(-1)
            error_queue = multiprocessing.Queue(-1)
            measuring: Synchronized = multiprocessing.Value("b", True)
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=len(valid_packet),
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            reader_job(
                queue=queue, error_queue=error_queue, measuring=measuring, args=args
            )

            assert not queue.empty()
            # エラーが発生していないことの確認
            assert error_queue.empty()
            records = queue.get()
            # 一回目の取得時は1秒分のレコードが入っている
            assert len(records) == args.record_per_sec
            records = queue.get()
            # 二回目の取得時はループを抜けているので残りが全て取得できる
            assert len(records) == 0
            records = queue.get()
            # ループを抜けたので終了フラグのNoneを返す
            assert records is None
            assert queue.empty()
            queue.close()
            error_queue.close()

    class TestWriterJob:
        def test_writer_job_success(self, tmpdir):
            queue = multiprocessing.Queue(-1)
            error_queue = multiprocessing.Queue(-1)
            args = WriterArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                logger_id="RP1",
                record_per_file=2,
            )
            data = [MeasureData(index=0, count=0, temperature=0, x=0, y=0, z=0, flag=0)]
            # 1件のデータを3回送る、1ファイルあたり2レコードだから2つファイルができる
            queue.put(data)
            queue.put(data)
            queue.put(data)
            queue.put(None)
            writer_job(
                queue=queue, error_queue=error_queue, args=args, output_dir=tmpdir
            )

            # queueから全部取得しているか
            assert queue.empty()
            # エラーが発生していないか
            assert error_queue.empty()
            # 2ファイル出力されているか
            csvs = sorted(os.listdir(tmpdir))
            assert csvs == ["0.csv", "1.csv"]
            queue.close()
            error_queue.close()

        def test_writer_job_empty_records_should_not_create_file(self, tmpdir):
            queue = multiprocessing.Queue(-1)
            error_queue = multiprocessing.Queue(-1)
            args = WriterArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                logger_id="RP1",
                record_per_file=2,
            )
            data = [MeasureData(index=0, count=0, temperature=0, x=0, y=0, z=0, flag=0)]
            # 1件のデータを2回と空の配列を送る、1ファイルあたり2レコードだから1つファイルができる
            queue.put(data)
            queue.put(data)
            queue.put([])
            queue.put(None)
            writer_job(
                queue=queue, error_queue=error_queue, args=args, output_dir=tmpdir
            )

            # queueから全部取得しているか
            assert queue.empty()
            # エラーが発生していないか
            assert error_queue.empty()
            # 2ファイル出力されているか
            csvs = sorted(os.listdir(tmpdir))
            assert csvs == ["0.csv"]
            queue.close()
            error_queue.close()

    class TestReaderJobArgs:
        def test_complement_data_success_when_prev_is_none(self):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            actual = args.complement_data(
                None,
                MeasureData(index=0, count=1, temperature=0, x=0, y=0, z=0, flag=0),
            )
            assert actual == [
                MeasureData(index=0, count=1, temperature=0, x=0, y=0, z=0, flag=0)
            ]

        def test_complement_data_success_when_no_missing(self):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            actual = args.complement_data(
                MeasureData(index=0, count=3, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=1, count=0, temperature=0, x=0, y=0, z=0, flag=0),
            )
            assert actual == [
                MeasureData(index=1, count=0, temperature=0, x=0, y=0, z=0, flag=0)
            ]

        def test_complement_data_success_when_3_data_missing(self):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            actual = args.complement_data(
                MeasureData(index=0, count=2, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=0, count=1, temperature=0, x=0, y=0, z=0, flag=0),
            )
            assert actual == [
                MeasureData(index=1, count=3, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=2, count=0, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=3, count=1, temperature=0, x=0, y=0, z=0, flag=0),
            ]

        def test_complement_data_success_when_first_data_but_missing(self):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            actual = args.complement_data(
                None,
                MeasureData(index=0, count=2, temperature=0, x=0, y=0, z=0, flag=0),
            )
            assert actual == [
                MeasureData(index=0, count=1, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=1, count=2, temperature=0, x=0, y=0, z=0, flag=0),
            ]

        def test_complement_data_success_count_start_is_not_zero(self):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=1,
                count_max=4,
                count_start=1,
            )
            actual = args.complement_data(
                None,
                MeasureData(index=0, count=2, temperature=0, x=0, y=0, z=0, flag=0),
            )
            assert actual == [
                MeasureData(index=0, count=1, temperature=0, x=0, y=0, z=0, flag=0),
                MeasureData(index=1, count=2, temperature=0, x=0, y=0, z=0, flag=0),
            ]

        @pytest.mark.parametrize(
            "count_start, count_diff",
            [
                (
                    (rms_pp_interval * 600 + 2) % int_16bit_max,
                    rms_pp_interval * 600,
                )
                for rms_pp_interval in range(1, 256)
            ],
        )
        def test_complement_data_success_count_max(self, count_start, count_diff):
            args = ReaderArgsForTest(
                model="TEST",
                serial="serial",
                port="/dev/AAA",
                baud=100,
                record_size=4,
                record_per_sec=2,
                record_begin=0x80,
                record_end=0x0D,
                start_command=[],
                end_command=[],
                count_diff=count_diff,
                count_max=int_16bit_max,
                count_start=count_start,
            )
            actual = args.complement_data(
                None,
                MeasureData(
                    index=0,
                    count=(count_start + count_diff * 9) % int_16bit_max,
                    temperature=0,
                    x=0,
                    y=0,
                    z=0,
                    flag=0,
                ),
            )
            # 2件補完されている
            assert len(actual) == 10
            for i in range(9):
                if actual[i + 1].count - actual[i].count < count_diff:
                    assert (actual[i].count + count_diff) % int_16bit_max == actual[
                        i + 1
                    ].count
                else:
                    assert actual[i + 1].count - actual[i].count == count_diff
